rm(list=ls())
library(lsl)
library(GPArotation)

#True Model Specification#
P <- 20
M <- 4
lambda0 <- diag(1, M) %x% matrix(c(.8, .65, .75, .65, .8), 5, 1)
lambda0[2, 2] <- lambda0[7, 3] <- lambda0[12, 4] <- lambda0[17, 1] <- .5
lambda0[9, 1] <- lambda0[14, 2] <- lambda0[19, 3] <- lambda0[4, 4] <- .5
phi0 <- 0.3 * matrix(1, M, 1) %*% matrix(1, 1, M) + diag( .7, M)
sigma0 <- lambda0 %*% phi0 %*% t(lambda0)
diag(sigma0) <- 1
psi0 <- sigma0 - lambda0 %*% phi0 %*% t(lambda0)


#Function for Generating Data#
generate_sample <- function(N, P, sigma) {
  Y <- matrix(rnorm(N * P), N, P)
  Y <- Y %*% chol(sigma)
  Y <- scale(Y, scale = F)
  return(as.data.frame(Y))
}

#Model Pattern for PL#
lambda_pl <- 1*(lambda0!=0)
lambda_pl[lambda0 < 0.6] <- NA
lambda_pl[1, 1] <- lambda_pl[6, 2] <- lambda_pl[11, 3] <- lambda_pl[16, 4] <- 0
phi_pl <- matrix(1, M, M)
nu_pml <- matrix(0, P, 1)
pattern_pl <- list(lambda = lambda_pl, phi = phi_pl, nu = nu_pml)

#Model Pattern for Oracle ML#
lambda_oml <- 1*(lambda0!=0)
lambda_oml[lambda0 < 0.2] <- NA
lambda_oml[1, 1] <- lambda_oml[6, 2] <- lambda_oml[11, 3] <- lambda_oml[16, 4] <- 0
phi_oml <- matrix(1, M, M)
nu_oml <- matrix(0, P, 1)
pattern_oml <- list(lambda = lambda_oml, phi = phi_oml, nu = nu_oml)

#Model Pattern for Non-Oracle ML#
lambda_nml <- matrix(1, P, M)
lambda_nml[1, 1] <- lambda_nml[6, 2] <- lambda_nml[11, 3] <- lambda_nml[16, 4] <- 0
phi_nml <- matrix(NA, M, M)
diag(phi_nml) <- 1
nu_nml <- matrix(0, P, 1)
pattern_nml <- list(lambda = lambda_nml, phi = phi_nml, nu = nu_nml)


#Starting Value Specification#
lambda_start <- diag(1, M) %x% matrix(rep(.8, 5), 5, 1)
value <- list(lambda = lambda_start)

#Penalty Specification#
penalty_l1 <- list(type = "l1", gamma = seq(.01, .3, .01), delta = 1)
penalty_scad <- list(type = "scad", gamma = seq(.01, .3, .01), delta = c(2.2, 2.6, 3))
penalty_mcp <- list(type = "mcp", gamma = seq(.01, .3, .01), delta = c(1.2, 1.6, 2))
penalty_oml <- list(type = "l1", gamma = 10000, delta = 1)
penalty_nml <- list(type = "l1", gamma = 10000, delta = 1)

#Convergence Criterion#
control <- list(maxit = 1000, epsilon = 10^-5)


#Simulation Start#
rpmax <- 700
N_all <- c(100, 200, 400, 800, 1600)

for (N in N_all) {
  set.seed(N)
  assign("tt_l1_aic", matrix(NA, 106, rpmax))
  assign("tt_l1_bic", matrix(NA, 106, rpmax))
  assign("tt_scad_aic", matrix(NA, 106, rpmax))
  assign("tt_scad_bic", matrix(NA, 106, rpmax))
  assign("tt_mcp_aic", matrix(NA, 106, rpmax))
  assign("tt_mcp_bic", matrix(NA, 106, rpmax))
  assign("tt_oml", matrix(NA, 106, rpmax))
  assign("tt_nml", matrix(NA, 106, rpmax))
  assign("itmax_all", matrix(NA, 5, rpmax))
  
  for (rp in 1:rpmax) {
    #Generating Data#
    Y <- generate_sample(N, P, sigma0)
    
    #Fit PL with L1 Penalty#
    rc_sem_l1 <- lslSEM()
    rc_sem_l1$input(raw = Y)
    rc_sem_l1$specify(pattern = pattern_pl, value = value, scale = F)
    rc_sem_l1$learn(penalty = penalty_l1, control = control)
    rst_l1 <- rc_sem_l1$summarize(type = "individual")
    tt_l1_aic[, rp] <- rst_l1[,1]
    tt_l1_bic[, rp] <- rst_l1[,2]
    itmax_all[1, rp] <- max(rc_sem_l1$structure$overall["it",,])
    
    #Fit PL with SCAD Penalty#
    rc_sem_scad <- lslSEM()
    rc_sem_scad$input(raw = Y)
    rc_sem_scad$specify(pattern = pattern_pl, value = value, scale = F)
    rc_sem_scad$learn(penalty = penalty_scad, control = control)
    rst_scad <- rc_sem_scad$summarize(type = "individual")
    tt_scad_aic[, rp] <- rst_scad[,1]
    tt_scad_bic[, rp] <- rst_scad[,2]
    itmax_all[2, rp] <- max(rc_sem_scad$structure$overall["it",,])
    
    #Fit PL with MCP Penalty#
    rc_sem_mcp <- lslSEM()
    rc_sem_mcp$input(raw = Y)
    rc_sem_mcp$specify(pattern = pattern_pl, value = value, scale = F)
    rc_sem_mcp$learn(penalty = penalty_mcp, control = control)
    rst_mcp <- rc_sem_mcp$summarize(type = "individual")
    tt_mcp_aic[, rp] <- rst_mcp[,1]
    tt_mcp_bic[, rp] <- rst_mcp[,2]
    itmax_all[3, rp] <- max(rc_sem_mcp$structure$overall["it",,])
    
    #Fit Oracle ML#
    rc_sem_oml <- lslSEM()
    rc_sem_oml$input(raw = Y)
    rc_sem_oml$specify(pattern = pattern_oml, value = value, scale = F)
    rc_sem_oml$learn(penalty = penalty_oml, control = control)
    rst_oml <- rc_sem_oml$summarize(type = "individual")
    tt_oml[, rp] <- rst_oml[,1]
    itmax_all[4, rp] <- max(rc_sem_oml$structure$overall["it",,])
    
    #Fit Non-Oracle ML#
    rc_sem_nml <- lslSEM()
    rc_sem_nml$input(raw = Y)
    rc_sem_nml$specify(pattern = pattern_nml, value = value, scale = F)
    rc_sem_nml$learn(penalty = penalty_nml, control = control)
    rst_nml <- rc_sem_nml$summarize(type = "individual")
    tt_nml[, rp] <- rst_nml[,1]
    itmax_all[5, rp] <- max(rc_sem_nml$structure$overall["it",,])
    
    if (rp %% 10 == 0) {
      write.table(tt_l1_aic, paste("tt_l1_aic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_l1_bic, paste("tt_l1_bic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_scad_aic, paste("tt_scad_aic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_scad_bic, paste("tt_scad_bic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_mcp_aic, paste("tt_mcp_aic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_mcp_bic, paste("tt_mcp_bic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_oml, paste("tt_oml_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_nml, paste("tt_nml_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(itmax_all, paste("itmax_all_", N, ".txt", sep = ""), col.names = F, row.names = F)
    }
  }
}

#True Parameter Vector#
tt0 <- c(lambda0[is.na(lambda_pl) | lambda_pl == 1], diag(psi0), phi0[lower.tri(phi0, diag = T)])

#Object Function for Partially Specified Target#
vgQ.L2m <- function(lambda){
  W <- lambda0 < 0.6
  Method <- "L2m"
  f <- sum(W * (lambda ^ 2))
  Gq <- W * 2 * lambda
  return(list(Gq = Gq, f = f, Method = Method))
}


#Calculate Simulation Results#
PCTM <- matrix(NA, 5, 8)
TDR <- matrix(NA, 5, 8)
FDR <- matrix(NA, 5, 8)
MSE <- matrix(NA, 5, 8)
SB <- matrix(NA, 5, 8)

rpmax <- 500

for (N in N_all) {
  itmax_all <- read.table(paste0("itmax_all_", N, ".txt"))
  itmax_idc <- as.logical(apply(itmax_all < 500, 2, prod))
  tt_l1_aic <- read.table(paste0("tt_l1_aic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_l1_bic <- read.table(paste0("tt_l1_bic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_scad_aic <- read.table(paste0("tt_scad_aic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_scad_bic <- read.table(paste0("tt_scad_bic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_mcp_aic <- read.table(paste0("tt_mcp_aic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_mcp_bic <- read.table(paste0("tt_mcp_bic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  
  tt_nml <- read.table(paste0("tt_nml","_",N,".txt"))[,itmax_idc][,1:rpmax]
  for (rp in 1:rpmax) {
    lambda_nml <- matrix(NA, P, M)
    lambda_nml[is.na(lambda_pl) | lambda_pl == 1] <- tt_nml[1:76, rp]
    lambda_nml[is.na(lambda_nml)] <- .8
    lambda_initial <- lambda_nml
    Qm <- sum((lambda_initial[lambda0 < 0.6]^2))
    
    for (l in 1:10){
      rst_rotation <- GPFoblq(lambda_initial, Tmat = matrix(rnorm(M * M), M, M) + diag(5, M), method = "L2m", maxit = 1000)
      if (sum(rst_rotation$loading[lambda0 < 0.6]^2) < Qm) {
        lambda_nml <- rst_rotation$loading
        phi_nml <- rst_rotation$Phi
        Qm <- sum(rst_rotation$loading[lambda0 < 0.6]^2)
      }
    }
    phi_nml <- diag(sign(colMeans(sign(lambda_nml) * (lambda0 > 0.6)))) %*% phi_nml %*% diag(sign(colMeans(sign(lambda_nml) * (lambda0 > 0.6))))
    lambda_nml <- lambda_nml %*% diag(sign(colMeans(sign(lambda_nml) * (lambda0 > 0.6))))
    Tr <- diag(0.8 / c(lambda_nml[1, 1], lambda_nml[6, 2], lambda_nml[11, 3], lambda_nml[16, 4]), 4)
    Tr_iv <- diag(c(lambda_nml[1, 1], lambda_nml[6, 2], lambda_nml[11, 3], lambda_nml[16, 4]) / 0.8, 4)
    lambda_nml <- lambda_nml %*% Tr
    phi_nml <- Tr_iv %*% phi_nml %*% Tr_iv 
    tt_nml[1:76, rp] <- lambda_nml[is.na(lambda_pl) | lambda_pl == 1]
    tt_nml[97:106, rp] <- phi_nml[lower.tri(phi_nml, diag = T)]
  }
  tt_oml <- read.table(paste0("tt_oml","_",N,".txt"))[,itmax_idc][,1:rpmax]
  
  TDR[which(N_all==N),1] <- mean(apply(tt_l1_aic[1:76,][tt0[1:76]!=0,]!=0,2,mean))
  TDR[which(N_all==N),2] <- mean(apply(tt_l1_bic[1:76,][tt0[1:76]!=0,]!=0,2,mean))
  TDR[which(N_all==N),3] <- mean(apply(tt_scad_aic[1:76,][tt0[1:76]!=0,]!=0,2,mean))
  TDR[which(N_all==N),4] <- mean(apply(tt_scad_bic[1:76,][tt0[1:76]!=0,]!=0,2,mean))
  TDR[which(N_all==N),5] <- mean(apply(tt_mcp_aic[1:76,][tt0[1:76]!=0,]!=0,2,mean))
  TDR[which(N_all==N),6] <- mean(apply(tt_mcp_bic[1:76,][tt0[1:76]!=0,]!=0,2,mean))
  TDR[which(N_all==N),7] <- mean(apply(tt_nml[1:76,][tt0[1:76]!=0,]!=0,2,mean))
  TDR[which(N_all==N),8] <- mean(apply(tt_oml[1:76,][tt0[1:76]!=0,]!=0,2,mean))
  
  FDR[which(N_all==N),1] <- mean(apply(tt_l1_aic[1:76,][tt0[1:76]==0,]!=0,2,mean))
  FDR[which(N_all==N),2] <- mean(apply(tt_l1_bic[1:76,][tt0[1:76]==0,]!=0,2,mean))
  FDR[which(N_all==N),3] <- mean(apply(tt_scad_aic[1:76,][tt0[1:76]==0,]!=0,2,mean))
  FDR[which(N_all==N),4] <- mean(apply(tt_scad_bic[1:76,][tt0[1:76]==0,]!=0,2,mean))
  FDR[which(N_all==N),5] <- mean(apply(tt_mcp_aic[1:76,][tt0[1:76]==0,]!=0,2,mean))
  FDR[which(N_all==N),6] <- mean(apply(tt_mcp_bic[1:76,][tt0[1:76]==0,]!=0,2,mean))
  FDR[which(N_all==N),7] <- mean(apply(tt_nml[1:76,][tt0[1:76]==0,]!=0,2,mean))
  FDR[which(N_all==N),8] <- mean(apply(tt_oml[1:76,][tt0[1:76]==0,]!=0,2,mean))
  
  MSE[which(N_all==N),1] <- mean(apply((tt_l1_aic-tt0)^2,2,sum))
  MSE[which(N_all==N),2] <- mean(apply((tt_l1_bic-tt0)^2,2,sum))  
  MSE[which(N_all==N),3] <- mean(apply((tt_scad_aic-tt0)^2,2,sum))
  MSE[which(N_all==N),4] <- mean(apply((tt_scad_bic-tt0)^2,2,sum))
  MSE[which(N_all==N),5] <- mean(apply((tt_mcp_aic-tt0)^2,2,sum))
  MSE[which(N_all==N),6] <- mean(apply((tt_mcp_bic-tt0)^2,2,sum))
  MSE[which(N_all==N),7] <- mean(apply((tt_nml-tt0)^2,2,sum))
  MSE[which(N_all==N),8] <- mean(apply((tt_oml-tt0)^2,2,sum))
  
  PCTM[which(N_all==N),1] <- mean(apply((tt_l1_aic!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),2] <- mean(apply((tt_l1_bic!=0) == (tt0!=0), 2, prod))  
  PCTM[which(N_all==N),3] <- mean(apply((tt_scad_aic!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),4] <- mean(apply((tt_scad_bic!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),5] <- mean(apply((tt_mcp_aic!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),6] <- mean(apply((tt_mcp_bic!=0) == (tt0!=0), 2, prod)) 
  PCTM[which(N_all==N),7] <- mean(apply((tt_nml!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),8] <- mean(apply((tt_oml!=0) == (tt0!=0), 2, prod))
  
  SB[which(N_all==N),1] <- sum((apply(tt_l1_aic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),2] <- sum((apply(tt_l1_bic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),3] <- sum((apply(tt_scad_aic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),4] <- sum((apply(tt_scad_bic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),5] <- sum((apply(tt_mcp_aic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),6] <- sum((apply(tt_mcp_bic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),7] <- sum((apply(tt_nml, 1, mean)-tt0)^2)  
  SB[which(N_all==N),8] <- sum((apply(tt_oml, 1, mean)-tt0)^2)
}

result <- rbind(PCTM,TDR,FDR,MSE,SB)
write.csv(result, "result.csv", row.names = F)