rm(list=ls())
library(lsl)

#True Model Specification#
P <- 27
M <- 9
lambda0 <- diag(1, 9) %x% matrix(c(.8,.75,.8), 3, 1)
beta0 <- matrix(0, 9, 9)
beta0[2, 1] <- beta0[3, 2] <- beta0[5, 4] <- beta0[6, 5] <- beta0[8, 7] <- beta0[9, 8] <- .45
beta0[4, 1] <- beta0[5, 2] <- beta0[6, 3] <- beta0[7, 4] <- beta0[8, 5] <- beta0[9, 6] <- .55
beta0_iv <- solve(diag(1, 9) - beta0)
phi0 <- diag(0, 9)
phi0[1, 1] <- 1
for (m in 2:9) {phi0[m, m] <- (1 - sum((beta0_iv[m,] ^ 2) * diag(phi0)))}
psi0 <- diag(c(.36, 0.4375, .36), 27)
sigma0 <- lambda0 %*% beta0_iv %*% phi0 %*% t(beta0_iv) %*% t(lambda0) + psi0

#Function for Generating Data#
generate_sample <- function(N, P, Sg0) {
  Y <- matrix(rnorm(N * P), N, P)
  Y <- Y %*% chol(Sg0)
  return(as.data.frame(Y))
}

#Model Pattern for PL#
lambda_pl <- (lambda0 != 0)
lambda_pl[1, 1] <- lambda_pl[4, 2] <- lambda_pl[7, 3] <- lambda_pl[10, 4] <- lambda_pl[13, 5] <- 0 
lambda_pl[16, 6] <- lambda_pl[19, 7] <- lambda_pl[22, 8] <- lambda_pl[25, 9] <-0
beta_pl <- matrix(0, 9, 9)
beta_pl[lower.tri(beta_pl)] <- NA
phi_pl <- diag(1, M)
nu_pl <- matrix(0, P, 1)
pattern_pl <- list(lambda = lambda_pl, beta = beta_pl, phi = phi_pl, nu = nu_pl)


#Model Pattern for Oracle ML#
lambda_oml <- lambda_pl
beta_oml <- matrix(0, 9, 9)
beta_oml[lower.tri(beta_oml)] <- NA
beta_oml[beta0 != 0] <- 1
phi_oml <- diag(1, M)
nu_oml <- matrix(0, P, 1)
pattern_oml <- list(lambda = lambda_oml, beta = beta_oml, phi = phi_oml, nu = nu_oml)

#Model Pattern for Oracle ML#
lambda_nml <- lambda_pl
beta_nml <- matrix(0, 9, 9)
beta_nml[lower.tri(beta_nml)] <- 1
phi_nml <- diag(1, M)
nu_nml <- matrix(0, P, 1)
pattern_nml <- list(lambda = lambda_nml, beta = beta_nml, phi = phi_nml, nu = nu_nml)

#Starting Value Specification#
lambda_start <- diag(1, M) %x% matrix(rep(.8, 3), 3, 1)
value <- list(lambda = lambda_start)

#Penalty Specification#
penalty_l1 <- list(type = "l1", gamma = seq(.01, .3, .01), delta = 1)
penalty_scad <- list(type = "scad", gamma = seq(.01, .3, .01), delta = c(2.2, 2.6, 3))
penalty_mcp <- list(type = "mcp", gamma = seq(.01, .3, .01), delta = c(1.2, 1.6, 2.0))
penalty_oml <- list(type = "l1", gamma = 10000, delta = 1)
penalty_nml <- list(type = "l1", gamma = 10000, delta = 1)

#Convergence Criterion#
control <- list(maxit = 1000, epsilon = 10^-5)

#Simulation Start#
rpmax <- 700
N_all <- c(100,200,400,800,1600)

for (N in N_all) {
  set.seed(N)
  assign("tt_l1_aic", matrix(NA, 90, rpmax))
  assign("tt_l1_bic", matrix(NA, 90, rpmax))
  assign("tt_scad_aic", matrix(NA, 90, rpmax))
  assign("tt_scad_bic", matrix(NA, 90, rpmax))
  assign("tt_mcp_aic", matrix(NA, 90, rpmax))
  assign("tt_mcp_bic", matrix(NA, 90, rpmax))
  assign("tt_oml", matrix(NA, 90, rpmax))
  assign("tt_nml", matrix(NA, 90, rpmax))
  assign("itmax_all", matrix(NA, 5, rpmax))
  
  for (rp in 1:rpmax) {
    #Generating Data#
    Y <- generate_sample(N, P, sigma0)
    
    #Fit PL with L1 Penalty#
    rc_sem_l1 <- lslSEM()
    rc_sem_l1$input(raw = Y)
    rc_sem_l1$specify(pattern = pattern_pl, value = value, scale = F)
    rc_sem_l1$learn(penalty = penalty_l1, control = control)
    rst_l1 <- rc_sem_l1$summarize(type = "individual")
    tt_l1_aic[, rp] <- rst_l1[,1]
    tt_l1_bic[, rp] <- rst_l1[,2]
    itmax_all[1, rp] <- max(rc_sem_l1$structure$overall["it",,])
    
    #Fit PL with SCAD Penalty#
    rc_sem_scad <- lslSEM()
    rc_sem_scad$input(raw = Y)
    rc_sem_scad$specify(pattern = pattern_pl, value = value, scale = F)
    rc_sem_scad$learn(penalty = penalty_scad, control = control)
    rst_scad <- rc_sem_scad$summarize(type = "individual")
    tt_scad_aic[, rp] <- rst_scad[,1]
    tt_scad_bic[, rp] <- rst_scad[,2]
    itmax_all[2, rp] <- max(rc_sem_scad$structure$overall["it",,])
    
    #Fit PL with MCP Penalty#
    rc_sem_mcp <- lslSEM()
    rc_sem_mcp$input(raw = Y)
    rc_sem_mcp$specify(pattern = pattern_pl, value = value, scale = F)
    rc_sem_mcp$learn(penalty = penalty_mcp, control = control)
    rst_mcp <- rc_sem_mcp$summarize(type = "individual")
    tt_mcp_aic[, rp] <- rst_mcp[,1]
    tt_mcp_bic[, rp] <- rst_mcp[,2]
    itmax_all[3, rp] <- max(rc_sem_mcp$structure$overall["it",,])
    
    #Fit Oracle ML#
    rc_sem_oml <- lslSEM()
    rc_sem_oml$input(raw = Y)
    rc_sem_oml$specify(pattern = pattern_oml, value = value, scale = F)
    rc_sem_oml$learn(penalty = penalty_oml, control = control)
    rst_oml <- rc_sem_oml$summarize(type = "individual")
    tt_oml[, rp] <- rst_oml[,1]
    itmax_all[4, rp] <- max(rc_sem_oml$structure$overall["it",,])
    
    #Fit Non-Oracle ML#
    rc_sem_nml <- lslSEM()
    rc_sem_nml$input(raw = Y)
    rc_sem_nml$specify(pattern = pattern_nml, value = value, scale = F)
    rc_sem_nml$learn(penalty = penalty_nml, control = control)
    rst_nml <- rc_sem_nml$summarize(type = "individual")
    tt_nml[, rp] <- rst_nml[,1]
    itmax_all[5, rp] <- max(rc_sem_nml$structure$overall["it",,])
    
    if (rp %% 10 == 0) {
      write.table(tt_l1_aic, paste("tt_l1_aic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_l1_bic, paste("tt_l1_bic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_scad_aic, paste("tt_scad_aic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_scad_bic, paste("tt_scad_bic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_mcp_aic, paste("tt_mcp_aic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_mcp_bic, paste("tt_mcp_bic_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_oml, paste("tt_oml_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(tt_nml, paste("tt_nml_", N, ".txt", sep = ""), col.names = F, row.names = F)
      write.table(itmax_all, paste("itmax_all_", N, ".txt", sep = ""), col.names = F, row.names = F)
    }
  }
}

#True Parameter Vector#
tt0 <- c(lambda0[lambda_pl == 1], diag(psi0), beta0[is.na(beta_pl) | beta_pl == 1], diag(phi0))

#Calculate Simulation Results#
PCTM <- matrix(NA, 5, 8)
TDR <- matrix(NA, 5, 8)
FDR <- matrix(NA, 5, 8)
MSE <- matrix(NA, 5, 8)
SB <- matrix(NA, 5, 8)

rpmax <- 500

for (N in N_all) {
  itmax_all <- read.table(paste0("itmax_all_", N, ".txt"))
  itmax_idc <- as.logical(apply(itmax_all < 500, 2, prod))
  tt_l1_aic <- read.table(paste0("tt_l1_aic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_l1_bic <- read.table(paste0("tt_l1_bic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_scad_aic <- read.table(paste0("tt_scad_aic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_scad_bic <- read.table(paste0("tt_scad_bic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_mcp_aic <- read.table(paste0("tt_mcp_aic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_mcp_bic <- read.table(paste0("tt_mcp_bic","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_nml <- read.table(paste0("tt_nml","_",N,".txt"))[,itmax_idc][,1:rpmax]
  tt_oml <- read.table(paste0("tt_oml","_",N,".txt"))[,itmax_idc][,1:rpmax]
  
  TDR[which(N_all==N),1] <- mean(apply(tt_l1_aic[46:81,][tt0[46:81]!=0,]!=0,2,mean))
  TDR[which(N_all==N),2] <- mean(apply(tt_l1_bic[46:81,][tt0[46:81]!=0,]!=0,2,mean))
  TDR[which(N_all==N),3] <- mean(apply(tt_scad_aic[46:81,][tt0[46:81]!=0,]!=0,2,mean))
  TDR[which(N_all==N),4] <- mean(apply(tt_scad_bic[46:81,][tt0[46:81]!=0,]!=0,2,mean))
  TDR[which(N_all==N),5] <- mean(apply(tt_mcp_aic[46:81,][tt0[46:81]!=0,]!=0,2,mean))
  TDR[which(N_all==N),6] <- mean(apply(tt_mcp_bic[46:81,][tt0[46:81]!=0,]!=0,2,mean))
  TDR[which(N_all==N),7] <- mean(apply(tt_nml[46:81,][tt0[46:81]!=0,]!=0,2,mean))
  TDR[which(N_all==N),8] <- mean(apply(tt_oml[46:81,][tt0[46:81]!=0,]!=0,2,mean))
  
  FDR[which(N_all==N),1] <- mean(apply(tt_l1_aic[46:81,][tt0[46:81]==0,]!=0,2,mean))
  FDR[which(N_all==N),2] <- mean(apply(tt_l1_bic[46:81,][tt0[46:81]==0,]!=0,2,mean))
  FDR[which(N_all==N),3] <- mean(apply(tt_scad_aic[46:81,][tt0[46:81]==0,]!=0,2,mean))
  FDR[which(N_all==N),4] <- mean(apply(tt_scad_bic[46:81,][tt0[46:81]==0,]!=0,2,mean))
  FDR[which(N_all==N),5] <- mean(apply(tt_mcp_aic[46:81,][tt0[46:81]==0,]!=0,2,mean))
  FDR[which(N_all==N),6] <- mean(apply(tt_mcp_bic[46:81,][tt0[46:81]==0,]!=0,2,mean))
  FDR[which(N_all==N),7] <- mean(apply(tt_nml[46:81,][tt0[46:81]==0,]!=0,2,mean))
  FDR[which(N_all==N),8] <- mean(apply(tt_oml[46:81,][tt0[46:81]==0,]!=0,2,mean))
  
  MSE[which(N_all==N),1] <- mean(apply((tt_l1_aic-tt0)^2,2,sum))
  MSE[which(N_all==N),2] <- mean(apply((tt_l1_bic-tt0)^2,2,sum))  
  MSE[which(N_all==N),3] <- mean(apply((tt_scad_aic-tt0)^2,2,sum))
  MSE[which(N_all==N),4] <- mean(apply((tt_scad_bic-tt0)^2,2,sum))
  MSE[which(N_all==N),5] <- mean(apply((tt_mcp_aic-tt0)^2,2,sum))
  MSE[which(N_all==N),6] <- mean(apply((tt_mcp_bic-tt0)^2,2,sum))
  MSE[which(N_all==N),7] <- mean(apply((tt_nml-tt0)^2,2,sum))
  MSE[which(N_all==N),8] <- mean(apply((tt_oml-tt0)^2,2,sum))
  
  PCTM[which(N_all==N),1] <- mean(apply((tt_l1_aic!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),2] <- mean(apply((tt_l1_bic!=0) == (tt0!=0), 2, prod))  
  PCTM[which(N_all==N),3] <- mean(apply((tt_scad_aic!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),4] <- mean(apply((tt_scad_bic!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),5] <- mean(apply((tt_mcp_aic!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),6] <- mean(apply((tt_mcp_bic!=0) == (tt0!=0), 2, prod)) 
  PCTM[which(N_all==N),7] <- mean(apply((tt_nml!=0) == (tt0!=0), 2, prod))
  PCTM[which(N_all==N),8] <- mean(apply((tt_oml!=0) == (tt0!=0), 2, prod))
  
  SB[which(N_all==N),1] <- sum((apply(tt_l1_aic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),2] <- sum((apply(tt_l1_bic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),3] <- sum((apply(tt_scad_aic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),4] <- sum((apply(tt_scad_bic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),5] <- sum((apply(tt_mcp_aic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),6] <- sum((apply(tt_mcp_bic, 1, mean)-tt0)^2)  
  SB[which(N_all==N),7] <- sum((apply(tt_nml, 1, mean)-tt0)^2)  
  SB[which(N_all==N),8] <- sum((apply(tt_oml, 1, mean)-tt0)^2)
}

result <- rbind(PCTM,TDR,FDR,MSE,SB)
write.csv(result,"result.csv",row.names = F)